/*
 * Copyright (C), 2000-2005 by the monit project group.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <config.h>

#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif

#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif

#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#ifdef HAVE_SYSLOG_H
#include <syslog.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#include "monitor.h"
#include "customlog.h"


/**
 *  Implementation of a logger that appends log messages to a file
 *  using a specifical user-defined format.
 *
 *  @author Jan-Henrik Haukeland, <hauk@tildeslash.com>
 *  @author Marco Ermini, <marco.ermini@gmail.com>
 *
 *  @file
 */


/* ------------------------------------------------------------- Definitions */


static FILE *CUSTOMLOG= NULL;
/*static char time_str[STRLEN];*/
static pthread_mutex_t customlog_mutex= PTHREAD_MUTEX_INITIALIZER;


/* -------------------------------------------------------------- Prototypes */


static int  open_customlog();
static int  handle_customlogformat(char *, char *);


/* ------------------------------------------------------------------ Public */


/**
 * Initialize the customlog and 'log' function
 * @return TRUE if the customlog was successfully initialized
 */
int customlog_init() {

  if(!Run.docustomlog) {
    return TRUE;
  }

  if(!open_customlog()) {
    return FALSE;
  }

  if (!Run.customlogformat) {
    setcustomlogformat ("{%d/%m/%Y, %T}t, %a, %s, %m, %d");
  }

  /* Register log_close to be
     called at program termination */
  atexit(customlog_close);

  /* Check the alarmids */
  //check_alarmids();

  return TRUE;
  
}


/**
 * Log a message to customlogfile
 * @param s The string to log
 */
void customlog_log(const char *s) {
  
  ASSERT(s);
  
  if(! Run.docustomlog) goto nocustomlog;

  if(CUSTOMLOG) {
    LOCK(customlog_mutex)
      fprintf(CUSTOMLOG, "%s\n", s);
    END_LOCK;
  }

  nocustomlog:
  LOCK(customlog_mutex)
    fprintf(stderr, "%s\n", s);
    fflush(stderr);
  END_LOCK;
  
}



/**
 * Close the customlog file
 */
void customlog_close() {
  
  if(CUSTOMLOG  && (0 != fclose(CUSTOMLOG))) {
    log("%s: Error closing the customlog file -- %s\n",
	prog, STRERROR);
  }
  pthread_mutex_destroy(&customlog_mutex);
  CUSTOMLOG= NULL;
  
}

/**
 * Handle customlog event
 */
void handle_customlog(Event_T E, Service_T S) {

  int n=0,m=0;
  int size= STRLEN;
  char *buf= xcalloc(sizeof(char), size);
  char *tmp= xcalloc(sizeof(char), size);
  char *tmp1= xcalloc(sizeof(char), size);
  struct tm *tm_now;
  time_t now;
 
  ASSERT(E);
  ASSERT(S);
  ASSERT(S->custommessage);

  /* We will handle only first passed event, recurrent passed events are
   * ignored. Failed events are handled each time. */
  if(!E->state_changed && !E->state)
  {
    return;
  }
 
  while (n <= strlen (Run.customlogformat)) {
	switch ((Run.customlogformat)[n]) {
	case '%':
		if (n+1 <= strlen (Run.customlogformat)) {
			switch ((Run.customlogformat)[n+1]) {
			/* % character */
			case '%':
				buf[m++]= '%';
				break;
                        /* alarm id */
                        case 'a':
                                sprintf (tmp, "%d", S->alarmid);
                                strcpy (&(buf[m]), tmp);
                                m+= strlen(tmp);
                                break;
                        /* event type description */
                        case 'd':
                                strcpy(tmp, Event_get_description(E));
                                strcpy (&(buf[m]), tmp);
                                m+= strlen(tmp);
                                break;
			/* service group name */
			case 'g':
				if (S->group) {
					strcpy (&(buf[m]), S->group);
					m+= strlen(S->group);
				} else {
					strcpy (&(buf[m]), "no group");
					m+= 8;
				}
				break;
                        /* customlog message */
                        case 'm':
                                strcpy (&(buf[m]), S->custommessage);
                                m+= strlen(S->custommessage);
                                break;
                        /* event message */
                        case 'M':
                                if(E->message) {
                                        strcpy (&(buf[m]), E->message);
                                        m+= strlen(E->message);
                                }
                                break;
			/* service name */
			case 'n':
				if (S->name) {
					strcpy (&(buf[m]), S->name);
					m+= strlen(S->name);
				} else {
					strcpy (&(buf[m]), "no name");
					m+= 7;
				}
				break;
			/* severity */
			case 's':
				/* if resumed, back to normal */
				if(!(E->state)) {
					strcpy(&(buf[m]), "Normal");
					m += 6;
				} else
				/* if timeout, it is fatal  */
					if (E->id == EVENT_TIMEOUT) {
					strcpy(&(buf[m]), "Fatal");
					m += 5;
					break;
				} else
				/* otherwise, it is the severity defined */
					{
					switch (S->severity) {
					case SEVERITY_NORMAL:
						strcpy(&(buf[m]), "Normal");
						m += 6;
						break;
					case SEVERITY_WARNING:
						strcpy(&(buf[m]), "Warning");
						m += 7;
						break;
					case SEVERITY_MINOR:
						strcpy(&(buf[m]), "Minor");
						m += 5;
						break;
					case SEVERITY_MAJOR:
						strcpy(&(buf[m]), "Major");
						m += 5;
						break;
					case SEVERITY_CRITICAL:
						strcpy(&(buf[m]), "Critical");
						m += 8;
						break;
					case SEVERITY_FATAL:
						strcpy(&(buf[m]), "Fatal");
						m += 5;
						break;
					};
				};
				break;
			/* status of the event (failed or passed) */
			case 'S':
				if(E->state)
					strcpy(&(buf[m]), "Failed");
				else
					strcpy(&(buf[m]), "Passed");
				m+= 6;
				break;
			/* time (standard monit format: "%Z %b %e %T") */
			case 't':
				sprintf (tmp, "%s", timestr());
				strcpy (&(buf[m]), tmp);
				m+= strlen(tmp);
				break;
			/* service type */
			case 'y':
				sprintf (tmp, "%d", S->type);
				strcpy (&(buf[m]), tmp);
				m+= strlen(tmp);
				break;
			case 'Y':
				switch (S->type) {
				case TYPE_DEVICE:
					strcpy(&(buf[m]), "Device");
					m+= 6;
					break;
				case TYPE_DIRECTORY:
					strcpy(&(buf[m]), "Directory");
					m+= 9;
					break;
				case TYPE_FILE:
					strcpy(&(buf[m]), "File");
					m+= 4;
					break;
				case TYPE_PROCESS:
					strcpy(&(buf[m]), "Process");
					m+= 7;
					break;
				case TYPE_HOST:
					strcpy(&(buf[m]), "Host");
					m+= 4;
					break;
				}
				break;
			default:
				buf[m++]= '%';
				buf[m++]= (Run.customlogformat)[n+1];
			}  /* switch ((Run.customlogformat)[n+1]) */
			n+= 2;
			break;
		}
		else
			buf[m++]= (Run.customlogformat)[n++];
		break;
		/* '{' characters specify a format. */
	case '{':
		if (strlen (&((Run.customlogformat)[n+2]))) {
			int k=n;

			time(&now);
			tm_now= localtime(&now);
			handle_customlogformat(tmp, &((Run.customlogformat)[n+1]));
			n+= strlen(tmp);
			if (strlen (&((Run.customlogformat)[n+2]))) {
				switch ((Run.customlogformat)[n+2]) {
				/* by now only time is supported */
				case 't':
					if (strftime(tmp1, STRLEN, tmp, tm_now) <= 0) {
						*tmp1= '\0';
					} else {
						strcpy (&(buf[m]), tmp1);
						m+= strlen(tmp1);
					}
					break;
				default:
					strncpy (&(buf[m]), &((Run.customlogformat)[k]), strlen(tmp)+3);
					m+= strlen(tmp);
					m+= 3;
				}
				n+= 3;
			} else {
				strncpy (&(buf[m]), &((Run.customlogformat)[k]), strlen(tmp)+3);
				m+= strlen(tmp)+3;
			}
		} else
			buf[m++]= (Run.customlogformat)[n++];
		break;
	default:
		buf[m++]= (Run.customlogformat)[n++];
	}
   }
   buf[m] = '\0';

   customlog_log(buf);
  
   FREE(buf);
   FREE(tmp);
   FREE(tmp1);
}

/**
 * Retrieves the format string in the custom log.
 * i.e. standard Monit time string have the format: {%Z %b %e %T}t
 */
static int handle_customlogformat(char *buffer, char *customformat) {

  int n=0;
  int l=0;

  ASSERT(customformat);
  l= strlen (customformat);

  while ((n <= l) && (customformat[n] != '}')) {
	buffer[n]= customformat[n];
	n++;
  }
  if (customformat[n] == '}') {
    buffer[n]= '\0';
    return n;
  } else
    buffer[0]= '\0';

  return 0;
}


/* ----------------------------------------------------------------- Private */


/**
 * Open a custom log file
 */
static int open_customlog() {
 
  umask(LOGMASK);
  if((CUSTOMLOG= fopen(Run.customlogfile,"a+")) == (FILE *)NULL) {
    log("%s: Error opening the custom log file '%s' for writing -- %s\n",
	prog, Run.logfile, STRERROR);
    return(FALSE);
  }
  /* Set logger in unbuffered mode */
  setvbuf(CUSTOMLOG, NULL, _IONBF, 0);

  return TRUE;
  
}


/*  
 * Reset the customlogformat if changed
 */ 
void setcustomlogformat(char *customlogformat) {
 
  if(Run.customlogformat) {
 
    if(IS(Run.customlogformat, customlogformat)) {

      FREE(customlogformat);
      return;

    } else {

      FREE(Run.customlogformat);

    }
  }

  Run.customlogformat= customlogformat;

}

